"""
FIXME: 感觉这个 component 设计有点问题，做的事情太多，得拆！
"""

from PyQt5.QtCore import Qt, QModelIndex

from .textlist import TextlistModel, TextlistView


class ProviderModel(object):
    """
    一个 ProviderModel 对应左边栏「我的音乐」下的一个项目

    FIXME: 或许不应该命名为 ProviderModel，而是
    """
    def __init__(self, name, icon='♬ ', desc='', on_click=None, **kwargs):
        self.name = name
        self.icon = icon
        self.desc = desc
        self._on_click = on_click


class ProvidersModel(TextlistModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._association = {}

    def assoc(self, provider_id, pm):
        self._association[provider_id] = pm
        self.add(provider_id)

    def get(self, provider_id):
        return self._association.get(provider_id)

    def remove(self, provider_id):
        if not self._association.pop(provider_id, None):
            self.remove(provider_id)

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        provider_id = self._items[row]
        provider = self._association[provider_id]
        if role == Qt.DisplayRole:
            return provider.icon + ' ' + provider.name
        if role == Qt.ToolTipRole:
            return provider.desc
        if role == Qt.UserRole:
            return provider
        return super().data(index, role)


class ProvidersView(TextlistView):
    def __init__(self, parent):
        super().__init__(parent)

        self.setMinimumHeight(80)
        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        provider = index.data(role=Qt.UserRole)
        if provider._on_click:
            provider._on_click()



# -*- coding: utf-8 -*-

import os
import configparser
import logging
import random
from itertools import chain

from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QColor

from .consts import THEMES_DIR, USER_THEMES_DIR


logger = logging.getLogger(__name__)


class ThemeManager(object):
    def __init__(self, app):
        super().__init__()
        self._app = app

        self.current_theme = None
        self._themes = []    # config file name (theme name)

    def choose(self, theme_name):
        '''
        :param theme: theme unique name
        '''
        def recursive_update(widget):
            if hasattr(widget, 'set_theme_style'):
                widget.set_theme_style()
            for child in widget.children():
                if isinstance(child, QWidget):
                    recursive_update(child)

        self.set_theme(theme_name)
        recursive_update(self._app)

    def scan(self, themes_dir=[THEMES_DIR, USER_THEMES_DIR]):
        '''themes directory'''
        self._themes = []
        for directory in themes_dir:
            files = os.listdir(directory)
            for f in files:
                f_name, f_ext = f.split('.')
                if f_ext == 'colorscheme':
                    self._themes.append(f_name)

    def list(self):
        '''show themes list

        :param theme: theme unique name
        :return: themes name list
        '''
        if not self._themes:
            self.scan()
        return self._themes

    def get_theme(self, theme_name):
        '''
        :param theme: unique theme name
        :return: `Theme` object
        '''
        pass

    def set_theme(self, theme_name):
        '''set current theme'''
        self.current_theme = Theme(theme_name)


class Theme(object):
    def __init__(self, config_name=None):
        self._config = configparser.ConfigParser()
        self.name = config_name

        self.read(config_name)

    def read(self, config_file):
        if config_file is not None:
            config_file_path = os.path.abspath(THEMES_DIR + '/' + config_file +
                                               '.colorscheme')
            if not os.path.exists(config_file_path):
                config_file_path = os.path.abspath(
                    USER_THEMES_DIR + '/' + config_file + '.colorscheme')
                print('........ %s ............' % config_file_path)
            config = self._config.read(config_file_path)
            if config:
                return True
        return False

    @property
    def background_light(self):
        color_section = self._config['Background']
        return self._parse_color_str(color_section['color'])

    @property
    def background(self):
        color_section = self._config['BackgroundIntense']
        return self._parse_color_str(color_section['color'])

    @property
    def foreground_light(self):
        color_section = self._config['Foreground']
        return self._parse_color_str(color_section['color'])

    @property
    def foreground(self):
        color_section = self._config['ForegroundIntense']
        return self._parse_color_str(color_section['color'])

    @property
    def color0_light(self):
        color_section = self._config['Color0']
        return self._parse_color_str(color_section['color'])

    @property
    def color0(self):
        color_section = self._config['Color0Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color1_light(self):
        color_section = self._config['Color1']
        return self._parse_color_str(color_section['color'])

    @property
    def color1(self):
        color_section = self._config['Color1Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color2_light(self):
        color_section = self._config['Color2']
        return self._parse_color_str(color_section['color'])

    @property
    def color2(self):
        color_section = self._config['Color2Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color3_light(self):
        color_section = self._config['Color3']
        return self._parse_color_str(color_section['color'])

    @property
    def color3(self):
        color_section = self._config['Color3Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color4_light(self):
        color_section = self._config['Color4']
        return self._parse_color_str(color_section['color'])

    @property
    def color4(self):
        color_section = self._config['Color4Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color5_light(self):
        color_section = self._config['Color5']
        return self._parse_color_str(color_section['color'])

    @property
    def color5(self):
        color_section = self._config['Color5Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color6_light(self):
        color_section = self._config['Color6']
        return self._parse_color_str(color_section['color'])

    @property
    def color6(self):
        color_section = self._config['Color6Intense']
        return self._parse_color_str(color_section['color'])

    @property
    def color7_light(self):
        color_section = self._config['Color7']
        return self._parse_color_str(color_section['color'])

    @property
    def color7(self):
        color_section = self._config['Color7Intense']
        return self._parse_color_str(color_section['color'])

    def random_color(self):
        '''return one of the eight*2 color'''
        attributes = self.__class__.__dict__.keys()
        color_attrs = []
        for attr in attributes:
            if attr.startswith('color') and not attr.startswith('color0'):
                color_attrs.append(attr)
        color_attr = random.choice(color_attrs)
        return getattr(self, color_attr)

    def _parse_color_str(self, color_str):
        rgb = [int(x) for x in color_str.split(',')]
        return QColor(rgb[0], rgb[1], rgb[2])


def get_colors_ctx(theme):
    return {
        'background': theme.background.name(),
        'background_light': theme.background_light.name(),
        'foreground': theme.foreground.name(),
        'foreground_light': theme.foreground_light.name(),
        'color0': theme.color0.name(),
        'color1': theme.color1.name(),
        'color2': theme.color2.name(),
        'color3': theme.color3.name(),
        'color4': theme.color4.name(),
        'color5': theme.color5.name(),
        'color6': theme.color6.name(),
        'color7': theme.color7.name(),
        'color0_light': theme.color0_light.name(),
        'color1_light': theme.color1_light.name(),
        'color2_light': theme.color2_light.name(),
        'color3_light': theme.color3_light.name(),
        'color4_light': theme.color4_light.name(),
        'color5_light': theme.color5_light.name(),
        'color6_light': theme.color6_light.name(),
        'color7_light': theme.color7_light.name(),
    }


def set_stylesheet(theme, widget):
    def do(w):
        object_name = w.objectName()
        ctx = get_colors_ctx(theme)
        ctx.update({'object_name': object_name})
        stylesheet = w.style_fmt.format(**ctx)
        w.setStyleSheet(stylesheet)

    discovered_w_list = []
    for w in chain([widget], discovered_w_list):
        if isinstance(w, QWidget):
            discovered_w_list += w.children()
            if hasattr(w, 'style_fmt'):
                do(w)



import asyncio
import logging
from PyQt5.QtCore import Qt, QTime, pyqtSignal, QTimer
from PyQt5.QtGui import QPainter, QKeySequence, QFontMetrics
from PyQt5.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSlider,
    QSplitter,
    QStyle,
    QVBoxLayout,
)

from fuocore.player import PlaybackMode, State

from feeluown.components.separator import Separator
from feeluown.components.playlists import PlaylistsView
from feeluown.components.provider import ProvidersView
from feeluown.components.history import HistoriesView
from feeluown.components.collections import CollectionsView
from feeluown.components.my_music import MyMusicView
from feeluown.components.volume_button import VolumeButton
from feeluown.containers.magicbox import MagicBox
from feeluown.containers.table_container import SongsTableContainer

from .helpers import use_mac_theme
from .utils import parse_ms

logger = logging.getLogger(__name__)


class ProgressSlider(QSlider):
    pause_player_needed = pyqtSignal()
    resume_player_needed = pyqtSignal()
    change_position_needed = pyqtSignal([int])

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setOrientation(Qt.Horizontal)
        self.sliderPressed.connect(self._on_pressed)
        self.sliderReleased.connect(self._on_released)

    def _on_pressed(self):
        self.pause_player_needed.emit()

    def _on_released(self):
        self.change_position_needed.emit(self.value())
        self.resume_player_needed.emit()

    def set_duration(self, ms):
        self.setRange(0, ms / 1000)

    def update_state(self, ms):
        self.setValue(ms / 1000)


class PlayerControlPanel(QFrame):

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        class IconButton(QPushButton):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                # 按钮文字一般是一个 symbol，长度控制为 40 是满足需求的
                self.setMaximumWidth(40)

        self._playback_modes = list(PlaybackMode.__members__.values())
        self._pm_alias_map = {
            PlaybackMode.one_loop: '单曲循环',
            PlaybackMode.sequential: '顺序播放',
            PlaybackMode.loop: '循环播放',
            PlaybackMode.random: '随机播放',
        }

        # initialize sub widgets
        self._layout = QHBoxLayout(self)
        self.previous_btn = IconButton(self)
        self.pp_btn = IconButton(self)
        self.next_btn = IconButton(self)

        #: playback mode switch button
        self.pms_btn = QPushButton(self)
        self.volume_btn = VolumeButton(self)
        self.volume_btn.change_volume_needed.connect(
            lambda volume: setattr(self._app.player, 'volume', volume))
        self.playlist_btn = IconButton(parent=self)

        #: mark song as favorite button
        self.like_btn = QPushButton(self)
        self.mv_btn = QPushButton('MV', self)
        self.download_btn = QPushButton(self)

        self.previous_btn.setObjectName('previous_btn')
        self.pp_btn.setObjectName('pp_btn')
        self.next_btn.setObjectName('next_btn')
        self.playlist_btn.setObjectName('playlist_btn')
        self.volume_btn.setObjectName('volume_btn')
        self.pms_btn.setObjectName('pms_btn')
        self.download_btn.setObjectName('download_btn')
        self.like_btn.setObjectName('like_btn')
        self.mv_btn.setObjectName('mv_btn')

        self.progress_slider = ProgressSlider(self)

        self.pms_btn.setToolTip('修改播放模式')
        self.volume_btn.setToolTip('调整音量')
        self.playlist_btn.setToolTip('显示当前播放列表')
        self.progress_slider.setToolTip('拖动调节进度')

        # TODO: implementation
        self.mv_btn.setToolTip('播放 MV（未实现，欢迎 PR）')
        self.download_btn.setToolTip('下载歌曲（未实现，欢迎 PR）')
        self.like_btn.setToolTip('收藏歌曲（未实现，欢迎 PR）')

        if not use_mac_theme():
            self.previous_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaSkipBackward))
            self.pp_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            self.next_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaSkipForward))
            self.volume_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaVolume))
            self.playlist_btn.setText('🎶')
        else:
            self.pp_btn.setCheckable(True)
            self.like_btn.setCheckable(True)
            self.download_btn.setCheckable(True)

        self.song_title_label = QLabel('No song is playing.', parent=self)
        self.song_source_label = QLabel('歌曲来源', parent=self)
        self.song_title_label.setAlignment(Qt.AlignCenter)
        self.duration_label = QLabel('00:00', parent=self)
        self.position_label = QLabel('00:00', parent=self)

        self.song_source_label.setObjectName('song_source_label')

        self.next_btn.clicked.connect(self._app.player.play_next)
        self.previous_btn.clicked.connect(self._app.player.play_previous)
        self.pp_btn.clicked.connect(self._app.player.toggle)
        self.pms_btn.clicked.connect(self._switch_playback_mode)

        self._app.player.state_changed.connect(self._on_player_state_changed)
        self._app.player.playlist.playback_mode_changed.connect(
            self.on_playback_mode_changed)
        self._app.player.playlist.song_changed.connect(
            self.on_player_song_changed)
        self.progress_slider.resume_player_needed.connect(self._app.player.resume)
        self.progress_slider.pause_player_needed.connect(self._app.player.pause)
        self.progress_slider.change_position_needed.connect(
            lambda value: setattr(self._app.player, 'position', value))

        self._update_pms_btn_text()
        self._setup_ui()

    def _setup_ui(self):
        # set widget layout
        self.progress_slider.setMinimumWidth(480)
        self.progress_slider.setMaximumWidth(600)
        self.song_source_label.setFixedHeight(20)
        self.progress_slider.setFixedHeight(20)  # half of parent height
        self.position_label.setFixedWidth(45)
        self.duration_label.setFixedWidth(45)
        self.like_btn.setFixedSize(15, 15)
        self.download_btn.setFixedSize(15, 15)
        self.mv_btn.setFixedHeight(16)

        self.progress_slider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self._sub_layout = QVBoxLayout()
        self._sub_top_layout = QHBoxLayout()

        # add space to make top layout align with progress slider
        self._sub_top_layout.addSpacing(3)
        self._sub_top_layout.addWidget(self.song_source_label)
        self._sub_top_layout.addSpacing(5)
        self._sub_top_layout.addWidget(self.song_title_label)
        self._sub_top_layout.addStretch(0)
        self._sub_top_layout.addWidget(self.like_btn)
        self._sub_top_layout.addSpacing(8)
        self._sub_top_layout.addWidget(self.mv_btn)
        self._sub_top_layout.addSpacing(8)
        self._sub_top_layout.addWidget(self.download_btn)
        self._sub_top_layout.addSpacing(3)

        self._sub_layout.addSpacing(3)
        self._sub_layout.addLayout(self._sub_top_layout)
        self._sub_layout.addWidget(self.progress_slider)

        self._layout.addSpacing(20)
        self._layout.addWidget(self.previous_btn)
        self._layout.addSpacing(8)
        self._layout.addWidget(self.pp_btn)
        self._layout.addSpacing(8)
        self._layout.addWidget(self.next_btn)
        self._layout.addSpacing(26)
        self._layout.addWidget(self.volume_btn)
        self._layout.addStretch(0)
        self._layout.addWidget(self.position_label)
        self._layout.addSpacing(7)
        self._layout.addLayout(self._sub_layout)
        self._layout.addSpacing(7)
        self._layout.addWidget(self.duration_label)
        self._layout.addSpacing(5)
        self._layout.addStretch(0)
        self._layout.addWidget(self.pms_btn)
        self._layout.addSpacing(8)
        self._layout.addWidget(self.playlist_btn)
        self._layout.addSpacing(18)

        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

    def _switch_playback_mode(self):
        playlist = self._app.player.playlist
        pm_total = len(self._playback_modes)
        pm_idx = self._playback_modes.index(playlist.playback_mode)
        if pm_idx < pm_total - 1:
            pm_idx += 1
        else:
            pm_idx = 0
        playlist.playback_mode = self._playback_modes[pm_idx]

    def on_duration_changed(self, duration):
        m, s = parse_ms(duration)
        t = QTime(0, m, s)
        self.duration_label.setText(t.toString('mm:ss'))

    def on_position_changed(self, position):
        m, s = parse_ms(position)
        t = QTime(0, m, s)
        self.position_label.setText(t.toString('mm:ss'))

    def on_playback_mode_changed(self, playback_mode):
        self._update_pms_btn_text()

    def _update_pms_btn_text(self):
        playback_mode = self._app.player.playlist.playback_mode
        alias = self._pm_alias_map[playback_mode]
        self.pms_btn.setText(alias)

    def on_player_song_changed(self, song):
        if song is None:
            self.song_source_label.setText('歌曲来源')
            self.song_title_label.setText('No song is playing.')
            return
        source_name_map = {p.identifier: p.name
                           for p in self._app.library.list()}
        font_metrics = QFontMetrics(QApplication.font())
        text = '{} - {}'.format(song.title, song.artists_name)
        elided_text = font_metrics.elidedText(
            text, Qt.ElideRight, self.progress_slider.width() - 100)
        self.song_source_label.setText(source_name_map[song.source])
        self.song_title_label.setText(elided_text)

    def _on_player_state_changed(self, state):
        self.pp_btn.setChecked(state == State.playing)


class TopPanel(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self._layout = QHBoxLayout(self)
        self.pc_panel = PlayerControlPanel(self._app, self)
        self.setObjectName('top_panel')

        self.setFixedHeight(60)

        self._layout.addWidget(self.pc_panel)


class LeftPanel(QFrame):

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.library_header = QLabel('音乐库', self)
        self.collections_header = QLabel('本地收藏 (Beta)', self)
        self.collections_header.setToolTip(
            '我们可以在本地建立『收藏集』来收藏自己喜欢的音乐资源\n\n'
            '每个收藏集都以一个独立 .fuo 文件的存在，'
            '将鼠标悬浮在收藏集上，可以查看文件所在路径。\n'
            '新建 fuo 文件，则可以新建收藏集，文件名即是收藏集的名字。\n\n'
            '手动编辑 fuo 文件即可编辑收藏集中的音乐资源，也可以在界面上拖拽来增删歌曲。'
        )
        self.playlists_header = QLabel('歌单列表', self)
        self.history_header = QLabel('浏览历史记录', self)
        self.my_music_header = QLabel('我的音乐', self)

        class Container(QFrame):
            def __init__(self, label, view, parent=None):
                super().__init__(parent)

                self._layout = QVBoxLayout(self)
                self._layout.setContentsMargins(0, 0, 0, 0)
                self._layout.setSpacing(0)
                label.setFixedHeight(25)
                self._layout.addWidget(label)
                self._layout.addWidget(view)
                self._layout.addStretch(0)
                # XXX: 本意是让 Container 下方不要出现多余的空间
                self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)


        self.playlists_view = PlaylistsView(self)
        self.providers_view = ProvidersView(self)
        self.histories_view = HistoriesView(self)
        self.my_music_view = MyMusicView(self)
        self.collections_view = CollectionsView(self)

        self.providers_con = Container(self.library_header, self.providers_view)
        self.collections_con = Container(self.collections_header, self.collections_view)
        self.histories_con = Container(self.history_header, self.histories_view)
        self.playlists_con = Container(self.playlists_header, self.playlists_view)
        self.my_music_con = Container(self.my_music_header, self.my_music_view)

        self.providers_view.setModel(self._app.providers)
        self.histories_view.setModel(self._app.histories)
        self.playlists_view.setModel(self._app.playlists)
        self.my_music_view.setModel(self._app.my_music)
        self.collections_view.setModel(self._app.collections)

        self._layout = QVBoxLayout(self)

        if use_mac_theme():
            self._layout.setSpacing(0)
            self._layout.setContentsMargins(6, 4, 0, 0)
        self._layout.addWidget(self.providers_con)
        self._layout.addWidget(self.collections_con)
        self._layout.addWidget(self.my_music_con)
        self._layout.addWidget(self.histories_con)
        self._layout.addWidget(self.playlists_con)
        self._layout.addStretch(0)

        self.providers_view.setFrameShape(QFrame.NoFrame)
        self.playlists_view.setFrameShape(QFrame.NoFrame)
        self.histories_view.setFrameShape(QFrame.NoFrame)
        self.my_music_view.setFrameShape(QFrame.NoFrame)
        self.collections_view.setFrameShape(QFrame.NoFrame)
        self.setMinimumWidth(180)
        self.setMaximumWidth(250)

        self.playlists_view.show_playlist.connect(
            lambda pl: asyncio.ensure_future(self.show_model(pl)))
        self.histories_view.show_model.connect(
            lambda model: asyncio.ensure_future(self.show_model(model)))
        self.collections_view.show_collection.connect(
            lambda collection: self._app.ui.table_container.show_collection(collection))

        # 让各个音乐库来决定是否显示这些组件
        self.playlists_con.hide()
        self.my_music_con.hide()

        # 历史记录暂时隐藏
        self.histories_con.hide()

    async def show_model(self, model):
        await self._app.ui.table_container.show_model(model)


class RightPanel(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self._layout = QHBoxLayout(self)
        self.table_container = SongsTableContainer(self._app, self)
        self._layout.addWidget(self.table_container)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)


class MessageLabel(QLabel):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.setObjectName('message_label')
        self._interval = 3
        self.timer = QTimer()
        self.queue = []
        self.hide()

        self.timer.timeout.connect(self.access_message_queue)

    @property
    def common_style(self):
        style_str = '''
            #{0} {{
                padding-left: 3px;
                padding-right: 5px;
            }}
        '''.format(self.objectName())
        return style_str

    def _set_error_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
            }}
        '''.format(self.objectName(),
                   theme.color1_light.name(),
                   theme.color7_light.name())
        self.setStyleSheet(style_str + self.common_style)

    def _set_normal_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
            }}
        '''.format(self.objectName(),
                   theme.color6_light.name(),
                   theme.color7.name())
        self.setStyleSheet(style_str + self.common_style)

    def show_message(self, text, error=False):
        if self.isVisible():
            self.queue.append({'error': error, 'message': text})
            self._interval = 1.5
            return
        if error:
            self._set_error_style()
        else:
            self._set_normal_style()
        self.setText(str(len(self.queue)) + ': ' + text)
        self.show()
        self.timer.start(self._interval * 1000)

    def access_message_queue(self):
        self.hide()
        if self.queue:
            m = self.queue.pop(0)
            self.show_message(m['message'], m['error'])
        else:
            self._interval = 3


class NetworkStatus(QLabel):
    def __init__(self, app, text=None, parent=None):
        super().__init__(text, parent)
        self._app = app

        self.setToolTip('这里显示的是当前网络状态')
        self.setObjectName('network_status_label')
        self._progress = 100
        self._show_progress = False

        self.set_state(1)

    def paintEvent(self, event):
        if self._show_progress:
            painter = QPainter(self)
            p_bg_color = self._app.theme_manager.current_theme.color0
            painter.fillRect(self.rect(), p_bg_color)
            bg_color = self._app.theme_manager.current_theme.color3
            rect = self.rect()
            percent = self._progress * 1.0 / 100
            rect.setWidth(int(rect.width() * percent))
            painter.fillRect(rect, bg_color)
            painter.drawText(self.rect(), Qt.AlignVCenter | Qt.AlignHCenter,
                             str(self._progress) + '%')
            self._show_progress = False
        else:
            super().paintEvent(event)

    @property
    def common_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
                padding-left: 5px;
                padding-right: 5px;
                font-size: 14px;
                font-weight: bold;
            }}
        '''.format(self.objectName(),
                   theme.color3.name(),
                   theme.background.name())
        return style_str

    def set_theme_style(self):
        self.setStyleSheet(self.common_style)

    def _set_error_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
            }}
        '''.format(self.objectName(),
                   theme.color5.name())
        self.setStyleSheet(self.common_style + style_str)

    def _set_normal_style(self):
        self.setStyleSheet(self.common_style)

    def set_state(self, state):
        if state == 0:
            self._set_error_style()
            self.setText('✕')
        elif state == 1:
            self._set_normal_style()
            self.setText('✓')

    def show_progress(self, progress):
        self._progress = progress
        self._show_progress = True
        if self._progress == 100:
            self._show_progress = False
        self.update()


class Ui:
    def __init__(self, app):
        self._app = app
        self._layout = QVBoxLayout(app)
        self._bottom_layout = QHBoxLayout()
        self._top_separator = Separator(parent=app)
        self._splitter = QSplitter(app)
        if use_mac_theme():
            self._splitter.setHandleWidth(0)

        # NOTE: 以位置命名的部件应该只用来组织界面布局，不要
        # 给其添加任何功能性的函数
        self.top_panel = TopPanel(app, app)
        self._left_panel_container = QScrollArea(self._app)
        self._left_panel_container.setWidgetResizable(True)
        self.left_panel = LeftPanel(self._app, self._splitter)
        self._left_panel_container.setWidget(self.left_panel)
        self.right_panel = RightPanel(self._app, self._splitter)

        # alias
        self.pc_panel = self.top_panel.pc_panel
        self.table_container = self.right_panel.table_container
        self.magicbox = MagicBox(self._app)

        # 对部件进行一些 UI 层面的初始化
        self._splitter.addWidget(self._left_panel_container)
        self._splitter.addWidget(self.right_panel)

        self.right_panel.setMinimumWidth(780)
        self._left_panel_container.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.right_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        if use_mac_theme():
            self._layout.addWidget(self.magicbox)
            self._layout.addWidget(self._splitter)
            self._layout.addWidget(self._top_separator)
            self._layout.addWidget(self.top_panel)
        else:
            self._layout.addWidget(self.top_panel)
            self._layout.addWidget(self._top_separator)
            self._layout.addWidget(self._splitter)
            self._layout.addWidget(self.magicbox)

        # self._layout.addLayout(self._bottom_layout)
        # self._bottom_layout.addWidget(self.magicbox)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.top_panel.layout().setSpacing(0)
        self.top_panel.layout().setContentsMargins(0, 0, 0, 0)

        self.pc_panel.playlist_btn.clicked.connect(self.show_player_playlist)

        self._app.hotkey_manager.registe(
            [QKeySequence('Ctrl+F'), QKeySequence(':'), QKeySequence('Alt+x')],
            self.magicbox.setFocus
        )

    def show_player_playlist(self):
        songs = self._app.playlist.list()
        self.table_container.show_player_playlist(songs)



import logging
import os
import random

from feeluown.consts import COLLECTIONS_DIR
from .parser import ModelParser
from .helpers import get_url


logger = logging.getLogger(__name__)


class Collection:

    def __init__(self, fpath, parser):
        self.fpath = fpath
        self._parser = parser

        self.name = ''
        self.models = []

    def load(self):
        """解析文件，初始化自己"""
        filepath = self.fpath
        filename = filepath.rsplit('/')[-1]
        name, _ = filename.split('.')
        self.name = name
        with open(filepath) as f:
            for line in f:
                model = self._parser.parse_line(line)
                if model is not None:
                    self.models.append(model)

    def add(self, song):
        if song not in self.models:
            line = self._parser.gen_line(song)
            with open(self.fpath, 'a+') as f:
                # 如果老的文件没有换行符，也要保证不会出错
                f.write('\n{}\n'.format(line))
            self.models.append(song)
        return True

    def remove(self, song):
        if song in self.models:
            url = get_url(song)
            with open(self.fpath, 'r+', encoding='utf-8') as f:
                lines = []
                for line in f:
                    if line.startswith(url):
                        continue
                    lines.append(line)
                f.seek(0)
                f.write(''.join(lines))
                # 确保最后写入一个换行符，让文件更加美观
                f.write('\n')
            self.models.remove(song)
        return True




class CollectionManager:
    def __init__(self, app):
        self._app = app
        self._library = app.library

        self.model_parser = ModelParser(self._library)

    def scan(self):
        directorys = self._app.config.COLLECTIONS_DIR + [COLLECTIONS_DIR]
        self._app.collections.clear()
        for directory in directorys:
            directory = os.path.expanduser(directory)
            for filename in os.listdir(directory):
                if not filename.endswith('.fuo'):
                    continue
                filepath = os.path.join(directory, filename)
                coll = Collection(filepath, self.model_parser)
                # TODO: 可以调整为并行
                coll.load()
                self._app.collections.add(coll)

        if not self._app.collections:
            fpath = self.gen_defualt_fuo()
            coll = Collection(fpath, self.model_parser)
            coll.load()
            self._app.collections.add(coll)

    @classmethod
    def gen_defualt_fuo(cls):
        logger.info('正在生成本地默认歌单')
        filenames = [
            '你真棒.fuo',
            '萌萌哒.fuo',
            '一起来 Hack.fuo',
            '今天也要元气满满喔.fuo',
            '正经.fuo',
            '有故事的歌曲.fuo',
        ]
        filename = random.choice(filenames)
        fpath = os.path.join(COLLECTIONS_DIR, filename)
        if not os.path.exists(fpath):
            with open(fpath, 'w') as f:
                lines = [
                    'fuo://netease/songs/16841667  # No Matter What - Boyzone',
                    'fuo://netease/songs/65800     # 最佳损友 - 陈奕迅',
                    'fuo://xiami/songs/3406085     # Love Story - Taylor Swift',
                    'fuo://netease/songs/5054926   # When You Say Noth… - Ronan Keating',
                    'fuo://qqmusic/songs/97773     # 晴天 - 周杰伦',
                    'fuo://qqmusic/songs/102422162 # 给我一首歌的时间 … - 周杰伦,蔡依林',
                    'fuo://xiami/songs/1769834090  # Flower Dance - DJ OKAWARI',
                ]
                f.write('\n'.join(lines))
        return fpath



import logging
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QAbstractItemView

from .textlist import TextlistModel, TextlistView


logger = logging.getLogger(__name__)


class CollectionsModel(TextlistModel):
    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDropEnabled
        return flags

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        item = self._items[row]
        if role == Qt.DisplayRole:
            return '◎  ' + item.name
        if role == Qt.ToolTipRole:
            return item.fpath
        return super().data(index, role)


class CollectionsView(TextlistView):
    show_collection = pyqtSignal([object])

    def __init__(self, parent):
        super().__init__(parent)
        self.setDragDropMode(QAbstractItemView.DropOnly)
        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        collection = index.data(role=Qt.UserRole)
        self.show_collection.emit(collection)

    # dragEnterEvent -> dragMoveEvent -> dropEvent
    def dragEnterEvent(self, e):
        """
        在这里仅判断此次 drag 的对象是否有效，如果无效，则忽略这个事件。
        之后 dropMoveEvent 也就不会接收到这个事件。
        """
        mimedata = e.mimeData()
        if mimedata.hasFormat('fuo-model/x-song'):
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        """
        由于鼠标移动时，可能会移动到非 collection item 的地方，
        所以在这里判断当前位置对应的 model index 是否符合条件。
        """
        # pylint: disable=all
        index = self.indexAt(e.pos())
        if index.isValid() and index.flags() & Qt.ItemIsDropEnabled:
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        mimedata = e.mimeData()
        assert mimedata.hasFormat('fuo-model/x-song')
        song = mimedata.model
        index = self.indexAt(e.pos())
        coll = index.data(Qt.UserRole)
        self._results[index.row] = (index, None)
        self.viewport().update()
        is_success = coll.add(song)
        self._results[index.row] = (index, is_success)
        self.viewport().update()
        self._result_timer.start(2000)
        e.accept()



"""
fuocore.protocol.handlers.helper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

良好的用文字展示一个对象

展示的时候需要注意以下几点：
1. 让 awk、grep 等 shell 工具容易处理

TODO: 让代码长得更好看
"""

from fuocore.models import ModelType


TYPE_NS_MAP = {
    ModelType.song: 'songs',
    ModelType.artist: 'artists',
    ModelType.album: 'albums',
    ModelType.playlist: 'playlists',
    ModelType.user: 'users',
    ModelType.lyric: 'lyrics',
}
URL_SCHEME = 'fuo'
NS_TYPE_MAP = {
    value: key
    for key, value in TYPE_NS_MAP.items()
}


def get_url(model):
    source = model.source
    ns = TYPE_NS_MAP[model.meta.model_type]
    identifier = model.identifier
    return 'fuo://{}/{}/{}'.format(source, ns, identifier)


def _fit_text(text, length, filling=True):
    """裁剪或者填补字符串，控制其显示的长度

    >>> _fit_text('12345', 6)
    '12345 '
    >>> _fit_text('哈哈哈哈哈s', 6)  # doctest: -ELLIPSIS
    '哈哈 …'
    >>> _fit_text('哈s哈哈哈哈s', 6)  # doctest: -ELLIPSIS
    '哈s哈…'
    >>> _fit_text('sssss', 5)
    'sssss'

    FIXME: 这样可能会截断一些英文词汇
    """
    assert 80 >= length >= 5

    text_len = 0
    len_index_map = {}
    for i, c in enumerate(text):
        # FIXME: 根据目前少量观察，在大部分字体下，
        # \u4e00 后面的字符宽度是英文字符两倍
        if ord(c) < 19968:  # ord(u'\u4e00')
            text_len += 1
            len_index_map[text_len] = i
        else:
            text_len += 2
            len_index_map[text_len] = i

    if text_len <= length:
        if filling:
            return text + (length - text_len) * ' '
        return text

    remain = length - 1
    if remain in len_index_map:
        return text[:(len_index_map[remain] + 1)] + '…'
    else:
        return text[:(len_index_map[remain - 1] + 1)] + ' …'


def show_song(song, uri_length=None, brief=False):
    """以一行文字的方式显示一首歌的信息

    :param uri_length: 控制 song uri 的长度
    :param brief: 是否只显示简要信息
    """
    artists_name = song.artists_name_display
    title = _fit_text(song.title_display, 18, filling=False)
    album_name = song.album_name_display

    song_uri = get_url(song)
    if uri_length is not None:
        song_uri = _fit_text(song_uri, uri_length)

    if brief:
        artists_name = _fit_text(artists_name, 20, filling=False)
        s = '{song}\t# {title} - {artists_name}'.format(
            song=song_uri,
            title=title,
            artists_name=artists_name)
        return s

    # XXX: 这个操作可能会产生网络请求
    album_uri = get_url(song.album)
    artists_uri = ','.join(get_url(artist) for artist in song.artists)
    msgs = (
        'provider     {}'.format(song.source),
        '     uri     {}'.format(song_uri),
        '   title     {}'.format(song.title),
        'duration     {}'.format(song.duration),
        '     url     {}'.format(song.url),
        ' artists     {}\t# {}'.format(artists_uri, artists_name),
        '   album     {}\t# {}'.format(album_uri, album_name),
    )
    return '\n'.join(msgs)


def show_songs(songs):
    uri_length = max((len(get_url(song)) for song in songs)) if songs else None
    return '\n'.join([show_song(song, uri_length=uri_length, brief=True)
                      for song in songs])


def show_artist(artist):
    msgs = [
        'provider      {}'.format(artist.source),
        'identifier    {}'.format(artist.identifier),
        'name          {}'.format(artist.name),
    ]
    if artist.songs:
        songs_header = ['songs::\n']
        songs = ['\t' + each for each in show_songs(artist.songs).split('\n')]
        msgs += songs_header
        msgs += songs
    return '\n'.join(msgs)


def show_album(album, brief=False):
    msgs = [
        'provider      {}'.format(album.source),
        'identifier    {}'.format(album.identifier),
        'name          {}'.format(album.name),
    ]
    if album.artists is not None:
        artists = album.artists
        artists_id = ','.join([str(artist.identifier) for artist in artists])
        artists_name = ','.join([artist.name for artist in artists])
        msgs_artists = ['artists       {}\t#{}'.format(artists_id, artists_name)]
        msgs += msgs_artists
    if not brief:
        msgs_songs_header = ['songs::\n']
        msgs_songs = ['\t' + each for each in show_songs(album.songs).split('\n')]
        msgs += msgs_songs_header
        msgs += msgs_songs
    return '\n'.join(msgs)


def show_playlist(playlist, brief=False):
    if brief:
        content = '{playlist}\t#{name}'.format(
            playlist=playlist,
            name=playlist.name)
    else:
        parts = [
            'name        {}'.format(playlist.name),
            'songs::\n',
        ]
        parts += ['\t' + show_song(song, brief=True) for song in playlist.songs]
        content = '\n'.join(parts)
    return content


def show_user(user):
    parts = [
        'name        {}'.format(user.name),
        'playlists::\n',
    ]
    parts += ['\t' + show_playlist(p, brief=True) for p in user.playlists]
    return '\n'.join(parts)



"""
fuocore.protocol.handlers.show
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

处理 ``show`` 命令::

    show fuo://               # 列出所有 provider
    show fuo://local/songs    # 显示本地所有歌曲
    show fuo://local/songs/1  # 显示一首歌的详细信息
"""
import logging
import re
from urllib.parse import urlparse

from .handlers import AbstractHandler
from .helpers import (
    show_song, show_artist, show_album, show_user,
    show_playlist
)

logger = logging.getLogger(__name__)


class NotFound(Exception):
    pass


class Router(object):
    rules = []
    handlers = {}

    @classmethod
    def register(cls, rule, handler):
        cls.rules.append(rule)
        cls.handlers[rule] = handler

    @classmethod
    def get_handler(cls, rule):
        return cls.handlers[rule]


def _validate_rule(rule):
    """简单的对 rule 进行校验

    TODO: 代码实现需要改进
    """
    if rule:
        if rule == '/':
            return
        parts = rule.split('/')
        if parts.count('') == 1:
            return
    raise ValueError('Invalid rule: {}'.format(rule))


def route(rule):
    """show handler router decorator

    example::

        @route('/<provider_name>/songs')
        def show_songs(provider_name):
            pass
    """
    _validate_rule(rule)

    def decorator(func):
        Router.register(rule, func)

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
        return wrapper
    return decorator


def regex_from_rule(rule):
    """为一个 rule 生成对应的正则表达式
    >>> regex_from_rule('/<provider>/songs')
    re.compile('^/(?P<provider>[^\\\/]+)/songs$')
    """
    kwargs_regex = re.compile(r'(<.*?>)')
    pattern = re.sub(
        kwargs_regex,
        lambda m: '(?P<{}>[^\/]+)'.format(m.group(0)[1:-1]),
        rule
    )
    regex = re.compile(r'^{}$'.format(pattern))
    return regex


def match(path, rules=Router.rules):
    """找到 path 对应的 rule，并解析其中的参数

    >>> match('/local/songs', rules=['/<p>/songs'])
    ('/<p>/songs', {'p': 'local'})

    :return: (rule, params) or None
    """
    for rule in rules:
        url_regex = regex_from_rule(rule)
        match = url_regex.match(path)
        if match:
            params = match.groupdict()
            return rule, params
    raise NotFound


def dispatch(req, rule, params):
    handler = Router.get_handler(rule)
    return handler(req, **params)


class ShowHandler(AbstractHandler):

    def handle(self, cmd):
        if cmd.args:
            furi = cmd.args[0]
        else:
            furi = 'fuo://'
        r = urlparse(furi)
        path = '/{}{}'.format(r.netloc, r.path)
        logger.debug('请求 path: {}'.format(path))
        try:
            rule, params = match(path)
        except NotFound as e:
            # FIXME: 抛一个合理的异常
            raise Exception('uri 不能被正确识别')
        return dispatch(self, rule, params)


@route('/')
def list_providers(req):
    provider_names = (provider.name for provider in
                      req.app.library.list())
    return '\n'.join(('fuo://' + name for name in provider_names))


@route('/<provider>/songs/<sid>')
def song_detail(req, provider, sid):
    provider = req.app.library.get(provider)
    song = provider.Song.get(sid)
    return show_song(song)


@route('/<provider>/songs/<sid>/lyric')
def lyric(req, provider, sid):
    provider = req.app.library.get(provider)
    song = provider.Song.get(sid)
    if song.lyric:
        return song.lyric.content
    return ''


@route('/<provider>/artists/<aid>')
def artist_detail(req, provider, aid):
    provider = req.app.library.get(provider)
    artist = provider.Artist.get(aid)
    return show_artist(artist)


@route('/<provider>/albums/<bid>')
def album_detail(req, provider, bid):
    provider = req.app.library.get(provider)
    album = provider.Album.get(bid)
    return show_album(album)


@route('/<provider>/users/<uid>')
def user_detail(req, provider, uid):
    provider = req.app.library.get(provider)
    user = provider.User.get(uid)
    return show_user(user)


@route('/<provider>/playlists/<pid>')
def playlist_detail(req, provider, pid):
    provider = req.app.library.get(provider)
    playlist = provider.Playlist.get(pid)
    return show_playlist(playlist)



class TipsManager(object):
    tips = [
        '你知道 mpris2 插件可以配合 osdlyrics 使用吗?',
        '双击歌手就可以浏览歌手的热门歌曲',
        '您可以自定义主题哦',
        '您有任何有关问题都可以上项目主页反馈哦',
        'Ctrl + F: 搜索快捷键, J, K 可以上下移动光标',
    ]

    def __init__(self, app):
        self._app = app

    def show_random_tip(self):
        pass



import hashlib
import logging

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import (
    QFormLayout,
    QLineEdit,
    QDialog,
    QPushButton,
    QLabel,
    QDialogButtonBox,
)

from fuocore.xiami.api import api
from fuocore.xiami.schemas import UserSchema

logger = logging.getLogger(__name__)


class LoginDialog(QDialog):
    """虾米音乐登录对话框"""

    login_success = pyqtSignal([object])

    def __init__(self, parent=None):
        super().__init__(parent)

        self._label = QLabel(self)
        self.username_input = QLineEdit(self)
        self.pw_input = QLineEdit(self)
        self.pw_input.setEchoMode(QLineEdit.Password)
        self._btn_box = QDialogButtonBox(self)
        self._ok_btn = QDialogButtonBox.Ok
        self._setup_ui()

        self.setWindowTitle('虾米账号密码登录')

        self._btn_box.clicked.connect(self.do_verify)

    def _setup_ui(self):
        self._btn_box.addButton(self._ok_btn)
        self._label.hide()

        self._layout = QFormLayout(self)
        self._layout.addRow('邮箱/手机号：', self.username_input)
        self._layout.addRow('密码：', self.pw_input)
        self._layout.addRow(self._label)
        self._layout.addRow(self._btn_box)

    def show_msg(self, msg, error=False):
        """显示提示信息"""
        self._label.show()
        self._label.setTextFormat(Qt.RichText)
        if error:
            color = 'red'
        else:
            color = 'green'
        self._label.setText('<span style="color: {};">{}</span>'
                            .format(color, msg))

    def do_verify(self):
        """校验用户名和密码，成功则发送信号"""
        username = self.username_input.text()
        password = self.pw_input.text()
        pw_md5digest = hashlib.md5(password.encode('utf-8')).hexdigest()
        rv = api.login(username, pw_md5digest)
        code, msg = rv['ret'][0].split('::')
        is_success = code == 'SUCCESS'
        self.show_msg(msg, error=(not is_success))
        if is_success:
            data = rv['data']['data']
            schema = UserSchema(strict=True)
            user, _ = schema.load(data)
            self.login_success.emit(user)
            self.close()



import asyncio
import logging

from fuocore.aio_tcp_server import TcpServer

from .handlers import (
    HelpHandler,
    SearchHandler,
    PlayerHandler,
    PlaylistHandler,
    StatusHandler,
)

from .show import ShowHandler  # noqa
from .parser import CmdParser  # noqa


logger = logging.getLogger(__name__)


def exec_cmd(app, live_lyric, cmd):
    # FIXME: 要么只传 app，要么把 player, live_lyric 等分别传进来
    # 目前更使用 app 作为参数

    logger.debug('EXEC_CMD: ' + str(cmd))

    # 一些
    if cmd.action in ('help', ):
        handler = HelpHandler(app,
                              live_lyric=live_lyric)

    elif cmd.action in ('show', ):
        handler = ShowHandler(app,
                              live_lyric=live_lyric)

    elif cmd.action in ('search', ):
        handler = SearchHandler(app,
                                live_lyric=live_lyric)

    # 播放器相关操作
    elif cmd.action in (
        'play', 'pause', 'resume', 'stop', 'toggle',
    ):
        handler = PlayerHandler(app,
                                live_lyric=live_lyric)

    # 播放列表相关命令
    elif cmd.action in (
        'add', 'remove', 'clear', 'list',
        'next', 'previous',
    ):
        """
        add/remove fuo://local:song:xxx
        create xxx

        set playback_mode=random
        set volume=100
        """
        handler = PlaylistHandler(app,
                                  live_lyric=live_lyric)
    elif cmd.action in ('status',):
        handler = StatusHandler(app,
                                live_lyric=live_lyric)
    else:
        return 'Oops Command not found!\n'

    rv = 'ACK {}'.format(cmd.action)
    if cmd.args:
        rv += ' {}'.format(' '.join(cmd.args))
    try:
        cmd_rv = handler.handle(cmd)
        if cmd_rv:
            rv += '\n' + cmd_rv
    except Exception as e:
        logger.exception('handle cmd({}) error'.format(cmd))
        return '\nOops\n'
    else:
        rv = rv or ''
        return rv + '\nOK\n'


class FuoProcotol:
    """fuo 控制协议

    TODO: 将这个类的实现移到另外一个模块，而不是放在 __init__.py 中
    TODO: 将命令的解析逻辑放在这个类里来实现
    """
    def __init__(self, app):
        self._app = app
        self._library = app.library
        self._live_lyric = app.live_lyric

    async def handle(self, conn, addr):
        app = self._app
        live_lyric = self._live_lyric

        event_loop = asyncio.get_event_loop()
        await event_loop.sock_sendall(conn, b'OK feeluown 1.0.0\n')
        while True:
            try:
                command = await event_loop.sock_recv(conn, 1024)
            except ConnectionResetError:
                logger.debug('客户端断开连接')
                break
            command = command.decode().strip()
            # NOTE: we will never recv empty byte unless
            # client close the connection
            if not command:
                conn.close()
                break
            logger.debug('RECV: %s', command)
            cmd = CmdParser.parse(command)
            msg = exec_cmd(app, live_lyric, cmd)
            await event_loop.sock_sendall(conn, bytes(msg, 'utf-8'))

    def run_server(self):
        port = 23333
        host = '0.0.0.0'
        event_loop = asyncio.get_event_loop()
        event_loop.create_task(
            TcpServer(host, port, handle_func=self.handle).run())
        logger.info('Fuo daemon run in {}:{}'.format(host, port))



# -*- coding: utf-8 -*-
import asyncio
import logging

from fuocore.player import MpvPlayer, Playlist as _Playlist


logger = logging.getLogger(__name__)


class Playlist(_Playlist):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

    @_Playlist.current_song.setter
    def current_song(self, song):
        """如果歌曲 url 无效，则尝试从其它平台找一个替代品"""
        if song is None or song.url:
            _Playlist.current_song.fset(self, song)
            return
        self.mark_as_bad(song)
        logger.info('Song:%s is invalid, try to get standby', song)
        songs = self._app.library.list_song_standby(song)
        if songs:
            song = songs[0]
        _Playlist.current_song.fset(self, song)


class Player(MpvPlayer):

    def __init__(self, app, *args, **kwargs):
        super().__init__(playlist=Playlist(app), *args, **kwargs)
        self._app = app
        self.initialize()



#! /usr/bin/env python3

import shutil
import os


def generate_icon():
    print('Generate icon, then you can see app in apps list.')
    os.system('mkdir -p ~/.FeelUOwn')
    DESKTOP_FILE = 'feeluown.desktop'
    current_path = os.path.abspath(os.path.dirname(__file__))
    icon = current_path + '/feeluown.png'
    feeluown_home_dir = os.path.expanduser('~') + '/.FeelUOwn'
    feeluown_icon = feeluown_home_dir + '/feeluown.png'
    shutil.copy(icon, feeluown_icon)

    icon_string = '#!/usr/bin/env xdg-open\n'\
                  '[Desktop Entry]\n'\
                  'Type=Application\n'\
                  'Name=FeelUOwn\n'\
                  'Comment=FeelUOwn Launcher\n'\
                  'Exec=python3 -m feeluown --log-to-file\n'\
                  'Icon={feeluown_icon}\n'\
                  'Categories=AudioVideo;Audio;Player;Qt;\n'\
                  'Terminal=false\n'\
                  'StartupNotify=true\n'\
                  .format(feeluown_icon=feeluown_icon)
    f_path = os.path.expanduser('~') +\
        '/.local/share/applications/' + DESKTOP_FILE
    try:
        with open(f_path, 'w') as f:
            f.write(icon_string)
        os.system('chmod +x %s' % f_path)
    except PermissionError:
        print('Please use sudo')
    en_desktop_path = os.path.expanduser('~') + '/Desktop'
    cn_desktop_path = os.path.expanduser('~') + '/桌面'
    if os.path.exists(en_desktop_path):
        desktop_f = en_desktop_path + '/' + DESKTOP_FILE
    if os.path.exists(cn_desktop_path):
        desktop_f = cn_desktop_path + '/' + DESKTOP_FILE
    shutil.copy(f_path, desktop_f)
    os.system('chmod +x %s' % desktop_f)
    print('Generate success.')


if __name__ == '__main__':
    generate_icon()



import logging

from abc import ABC, abstractmethod
from collections import defaultdict
from urllib.parse import urlparse

from fuocore.player import PlaybackMode, State
from .helpers import show_songs, show_song, get_url
from .parser import ModelParser


logger = logging.getLogger(__name__)


class AbstractHandler(ABC):
    def __init__(self, app, live_lyric):
        self.app = app
        self.live_lyric = live_lyric
        self.model_parser = ModelParser(self.app.library)

    @abstractmethod
    def handle(self, cmd):
        pass


class SearchHandler(AbstractHandler):
    def handle(self, cmd):
        return self.search_songs(cmd.args[0])

    def search_songs(self, query):
        logger.debug('搜索 %s ...', query)
        providers = self.app.library.list()
        source_in = [provd.identifier for provd in providers
                     if provd.Song.meta.allow_get]
        songs = []
        for result in self.app.library.search(query, source_in=source_in):
            logger.debug('从 %s 搜索到 %d 首歌曲，取前 20 首',
                         result.source, len(result.songs))
            songs.extend(result.songs[:20])
        logger.debug('总共搜索到 %d 首歌曲', len(songs))
        return show_songs(songs)


class StatusHandler(AbstractHandler):
    def handle(self, cmd):
        player = self.app.player
        playlist = self.app.playlist
        live_lyric = self.live_lyric
        repeat = int(playlist.playback_mode in
                     (PlaybackMode.one_loop, PlaybackMode.loop))
        random = int(playlist.playback_mode == PlaybackMode.random)
        msgs = [
            'repeat:    {}'.format(repeat),
            'random:    {}'.format(random),
            'volume:    {}'.format(player.volume),
            'state:     {}'.format(player.state.name),
        ]
        if player.state in (State.paused, State.playing):
            msgs += [
                'duration:  {}'.format(player.duration),
                'position:  {}'.format(player.position),
                'song:      {}'.format(show_song(player.current_song, brief=True)),  # noqa
                'lyric-s:   {}'.format(live_lyric.current_sentence),
            ]
        return '\n'.join(msgs)


class PlayerHandler(AbstractHandler):
    def handle(self, cmd):
        if cmd.action == 'play':
            song_furi = cmd.args[0]
            return self.play_song(song_furi.strip())
        elif cmd.action == 'pause':
            # FIXME: please follow ``Law of Demeter``
            self.app.player.pause()
        elif cmd.action == 'stop':
            self.app.player.stop()
        elif cmd.action == 'resume':
            self.app.player.resume()
        elif cmd.action == 'toggle':
            self.app.player.toggle()

    def play_song(self, song_furi):
        song = self.model_parser.parse_line(song_furi)
        if song is not None:
            self.app.player.play_song(song)


class PlaylistHandler(AbstractHandler):
    def handle(self, cmd):
        if cmd.action == 'add':
            return self.add(cmd.args[0].strip())
        elif cmd.action == 'remove':
            return self.remove(cmd.args[0].strip())
        elif cmd.action == 'clear':
            return self.clear()
        elif cmd.action == 'list':
            return self.list()
        elif cmd.action == 'next':
            self.app.player.play_next()
        elif cmd.action == 'previous':
            self.app.player.play_previous()

    def add(self, furis):
        playlist = self.app.playlist
        furi_list = furis.split(',')
        for furi in furi_list:
            song = self.model_parser.parse_line(furi)
            playlist.add(song)

    def remove(self, song_uri):
        # FIXME: a little bit tricky
        for song in self.app.playlist.list():
            if get_url(song) == song_uri:
                self.app.playlist.remove(song)
                break

    def list(self):
        songs = self.app.playlist.list()
        return show_songs(songs)

    def clear(self):
        self.app.playlist.clear()


class HelpHandler(AbstractHandler):
    def handle(self, cmd):
        return """
Available commands::

    search <string>  # search songs by <string>
    show fuo://xxx  # show xxx detail info
    play fuo://xxx/songs/yyy  # play yyy song
    list  # show player current playlist
    status  # show player status
    next  # play next song
    previous  # play previous song
    pause
    resume
    toggle

Watch live lyric::

    echo "sub topic.live_lyric" | nc host 23334
"""



# -*- coding: utf-8 -*-

import asyncio
import logging
from functools import partial

from fuocore.local import provider

from feeluown.app import App

__alias__ = '本地音乐'
__feeluown_version__ = '1.1.0'
__version__ = '0.1a0'
__desc__ = '本地音乐'

logger = logging.getLogger(__name__)


def show_provider(app):
    app.playlists.clear()
    # app.playlists.add(provider.playlists)

    app.ui.left_panel.my_music_con.hide()
    app.ui.left_panel.playlists_con.hide()
    app.ui.table_container.show_songs(provider.songs)


def enable(app):
    logger.info('Register provider: %s', provider)
    loop = asyncio.get_event_loop()
    future_scan = loop.run_in_executor(None, provider.scan)
    app.library.register(provider)
    if app.mode & App.GuiMode:
        from feeluown.components.provider import ProviderModel

        pm = ProviderModel(
            name='本地音乐',
            icon='♪ ',
            desc='点击显示所有本地音乐',
            on_click=partial(show_provider, app),
        )
        logger.info('Associate %s with %s', provider, pm.name)
        app.providers.assoc(provider.identifier, pm)
        future_scan.add_done_callback(lambda _: show_provider(app))


def disable(app):
    logger.info('唔，不要禁用我')



# -*- coding: utf-8 -*-

import json
import logging
import os

from fuocore.xiami.provider import provider
from fuocore.xiami.models import XUserModel

from feeluown.consts import DATA_DIR

__alias__ = '虾米音乐'
__version__ = '0.1a1'
__desc__ = '虾米音乐'

logger = logging.getLogger(__name__)
USER_INFO_FILE = DATA_DIR + '/xiami_user_info.json'


def dump_user(user):
    assert user.access_token is not None
    data = {
        'id': user.identifier,
        'name': user.name,
        'access_token': user.access_token,
    }
    with open(USER_INFO_FILE, 'w') as f:
        json.dump(data, f)


def load_user():
    if not os.path.exists(USER_INFO_FILE):
        return None
    with open(USER_INFO_FILE) as f:
        user_data = json.load(f)
    user = XUserModel(source=provider.identifier,
                      identifier=user_data['id'],
                      name=user_data['name'],
                      access_token=user_data['access_token'])
    return user


class Xiami(object):
    """GUI 控制"""

    def __init__(self, app):
        from feeluown.components.provider import ProviderModel

        self._app = app
        self._user = None

        self._pm = ProviderModel(
            name='虾米音乐',
            icon='♩ ',
            desc='',
            on_click=self.show_provider)
        self._app.providers.assoc(provider.identifier, self._pm)

    def show_login_dialog(self):
        from .ui import LoginDialog
        dialog = LoginDialog()
        dialog.login_success.connect(self.bind_user)
        dialog.exec()

    def bind_user(self, user, dump=True):
        if dump:
            dump_user(user)
        self._user = user
        provider.auth(user)

    def show_provider(self):
        """展示虾米首页

        要求用户已经登录，支持展示用户名、用户歌单列表

        TODO: 可以考虑支持展示榜单等
        """
        from feeluown.components.my_music import MyMusicItem

        if self._user is None:
            user = load_user()
            if user is None:
                self.show_login_dialog()
            else:
                self.bind_user(user, dump=False)
        if self._user is not None:
            # 显示用户名
            self._pm.name = '虾米音乐 - {}'.format(self._user.name)
            # 显示播放列表/歌单
            self._app.playlists.clear()
            self._app.playlists.add(self._user.playlists)
            self._app.playlists.add(self._user.fav_playlists, is_fav=True)
            # 显示用户收藏的歌曲
            self._app.ui.left_panel.my_music_con.show()
            self._app.ui.left_panel.playlists_con.show()
            self._app.my_music.clear()
            def func():
                self._app.ui.table_container.show_songs(self._user.fav_songs)
            self._app.my_music.add(MyMusicItem(
                '♥ 我的收藏', on_click=func))


def enable(app):
    app.library.register(provider)
    if app.mode & app.GuiMode:
        Xiami(app)


def disable(app):
    app.library.deregister(provider)



import logging

logger = logging.getLogger(__name__)


class PlayerModeManager(object):
    current_mode = None

    def __init__(self, app):
        super().__init__()
        self._app = app

    def enter_mode(self, mode):
        self.current_mode = mode
        self._app.player.change_player_mode_to_other()
        msg = '进入 %s 播放模式' % mode.name
        self._app.message(msg)
        logger.info(msg)
        try:
            mode.load()
        except Exception as e:
            logger.error('enter mode %s failed ' % mode.name)
            logger.error(str(e))

    def exit_to_normal(self):
        if self.current_mode is not None:
            self._app.player.change_player_mode_to_normal()
            try:
                self.current_mode.unload()
            except Exception as e:
                logger.error('exit mode %s failed ' % self.current_mode.name)
                logger.error(str(e))
            self.current_mode = None


class PlayerModeBase(object):
    def __init__(self, app):
        super().__init__()
        self._app = app
        self.player = self._app.player
        self.player.signal_playlist_finished.connect(
            self.on_playlist_finished)

    @property
    def name(self):
        raise NotImplementedError('This should be implemented by subclass')

    def on_playlist_finished(self):
        raise NotImplementedError('This should be implemented by subclass')

    def load(self):
        raise NotImplementedError('This should be implemented by subclass')

    def unload(self):
        self._app.message('退出 %s 播放模式' % self.name)
        self.player.signal_playlist_finished.disconnect(
            self.on_playlist_finished)



"""
feeluown.cliapp
~~~~~~~~~~~~~~~

It provides a CliMixin to make integration easilly.
"""


import logging

from fuocore.live_lyric import LiveLyric


logger = logging.getLogger(__name__)


class LiveLyricPublisher(object):
    topic = 'topic.live_lyric'

    def __init__(self, gateway):
        self.gateway = gateway
        gateway.add_topic(self.topic)

    def publish(self, sentence):
        self.gateway.publish(sentence + '\n', self.topic)


class CliAppMixin(object):
    """
    FIXME: Subclass must call init to make this mixin
    work properly, which seems to be little bit strange. But
    this mixin helps avoid duplicate code temporarily.
    """
    def __init__(self):
        live_lyric = LiveLyric()
        live_lyric_publisher = LiveLyricPublisher(self.pubsub_gateway)

        self.live_lyric = live_lyric
        self._live_lyric_publisher = live_lyric_publisher

        live_lyric.sentence_changed.connect(live_lyric_publisher.publish)
        self.player.position_changed.connect(live_lyric.on_position_changed)
        self.playlist.song_changed.connect(live_lyric.on_song_changed)



#!/usr/bin/env python3

from setuptools import setup

import feeluown


setup(
    name='feeluown',
    version=feeluown.__version__,
    description='*nix music player',
    author='Cosven',
    author_email='yinshaowen241@gmail.com',
    packages=[
        'feeluown',
        'feeluown.containers',
        'feeluown.components',
        'feeluown.plugins.qqmusic',
        'feeluown.plugins.neteasemusic',
        'feeluown.plugins.local',
        'feeluown.plugins.xiami',
        'feeluown.protocol',
        ],
    package_data={
        '': ['*.qss', '*.png', '../icons/*.png']
        },
    url='https://github.com/cosven/FeelUOwn',
    keywords=['media', 'player', 'application', 'PyQt5', 'Python 3'],
    classifiers=(
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Environment :: X11 Applications :: Qt',
        ),
    # FIXME depends on PyQt5
    install_requires=[
        'quamash>=0.5.5',
        'fuocore>=2.3',
        'requests',
        ],
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
                "feeluown=feeluown.__main__:main",
                "feeluown-genicon=feeluown.install:generate_icon",
                # "feeluown-update=feeluown.install:update"
            ]
        },
    )



import asyncio
import logging

from PyQt5.QtCore import Qt, QSize, pyqtSignal
from PyQt5.QtGui import QImage, QPixmap, QFont, QPalette, QPainter, QColor
from PyQt5.QtWidgets import (
    QDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QTabBar,
    QVBoxLayout,
    QWidget,
)

from fuocore import ModelType
from feeluown.helpers import use_mac_theme
from feeluown.components.songs_table import SongsTableModel, SongsTableView


logger = logging.getLogger(__name__)


class DescriptionContainer(QScrollArea):

    space_pressed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self._label = QLabel(self)
        self._label.setWordWrap(True)
        self._label.setTextFormat(Qt.RichText)
        self._label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.setWidget(self._label)

        self._setup_ui()

    def _setup_ui(self):
        self._label.setAlignment(Qt.AlignTop)
        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

    @property
    def html(self):
        return self._label.text()

    def set_html(self, desc):
        self._label.setText(desc)

    def keyPressEvent(self, event):
        key_code = event.key()
        if key_code == Qt.Key_J:
            value = self.verticalScrollBar().value()
            self.verticalScrollBar().setValue(value + 20)
        elif key_code == Qt.Key_K:
            value = self.verticalScrollBar().value()
            self.verticalScrollBar().setValue(value - 20)
        elif key_code == Qt.Key_Space:
            self.space_pressed.emit()
            event.accept()
        else:
            super().keyPressEvent(event)



class TableToolbar(QWidget):
    _desc_btn_checked_text = '折叠'
    _desc_btn_unchecked_text = '展开描述'

    play_all_needed = pyqtSignal()
    toggle_desc_needed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.play_all_btn = QPushButton('播放全部', self)
        self.desc_btn = QPushButton(self._desc_btn_unchecked_text, self)
        self.play_all_btn.clicked.connect(self.play_all_needed.emit)
        self.desc_btn.clicked.connect(self.on_desc_btn_toggled)
        self._setup_ui()

    def _setup_ui(self):
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.play_all_btn)
        self._layout.addStretch(1)
        self._layout.addWidget(self.desc_btn)
        self._layout.addStretch(0)

    def on_desc_btn_toggled(self, checked):
        if checked:
            self.play_all_btn.hide()
            self.desc_btn.setText(self._desc_btn_checked_text)
        else:
            self.play_all_btn.show()
            self.desc_btn.setText(self._desc_btn_unchecked_text)
        self.toggle_desc_needed.emit()


class SongsTableContainer(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.songs_table = SongsTableView(self)
        self._toolbar = TableToolbar(self)
        self._cover_label = QLabel(self)
        self._desc_container_folded = True
        self._desc_container = DescriptionContainer(self)
        self._top_container = QWidget(self)
        self._cover_container = QWidget(self._top_container)

        self.songs_table.play_song_needed.connect(
            lambda song: asyncio.ensure_future(self.play_song(song)))
        self.songs_table.show_artist_needed.connect(
            lambda artist: asyncio.ensure_future(self.show_model(artist)))
        self.songs_table.show_album_needed.connect(
            lambda album: asyncio.ensure_future(self.show_model(album)))

        self._desc_container.space_pressed.connect(self.toggle_desc_container_fold)
        self._toolbar.toggle_desc_needed.connect(self.toggle_desc_container_fold)
        self._toolbar.play_all_needed.connect(self.play_all)

        self.hide()
        self._setup_ui()

    def _setup_ui(self):
        self._left_sub_layout = QVBoxLayout(self._cover_container)
        self._top_layout = QHBoxLayout(self._top_container)
        self._layout = QVBoxLayout(self)

        self._cover_label.setMinimumWidth(200)
        self._right_sub_layout = QVBoxLayout()
        self._right_sub_layout.addWidget(self._desc_container)
        self._right_sub_layout.addWidget(self._toolbar)
        self._left_sub_layout.addWidget(self._cover_label)
        self._left_sub_layout.addStretch(0)
        # 根据 Qt 文档中所说，在大部分平台中，ContentMargin 为 11
        self._left_sub_layout.setContentsMargins(0, 0, 11, 0)
        self._left_sub_layout.setSpacing(0)

        self._top_layout.addWidget(self._cover_container)
        self._top_layout.addLayout(self._right_sub_layout)
        self._top_layout.setStretch(1, 1)

        self.setAutoFillBackground(False)
        if use_mac_theme():
            self._layout.setContentsMargins(0, 0, 0, 0)
            self._layout.setSpacing(0)
        self._layout.addWidget(self._top_container)
        self._layout.addWidget(self.songs_table)

        self._top_container.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        # FIXME: 更好的计算宽度和高度
        # 目前是假设知道自己初始化高度大约是 530px
        # 之后可以考虑按比例来计算
        self.overview_height = 180
        self._top_container.setMaximumHeight(self.overview_height)
        self._songs_table_height = 530 - self.overview_height
        self.songs_table.setMinimumHeight(self._songs_table_height)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    async def play_song(self, song):
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, lambda: song.url)
        self._app.player.play_song(song)

    def play_all(self):
        songs = self.songs_table.model().songs
        self._app.player.playlist.clear()
        for song in songs:
            self._app.player.playlist.add(song)
        self._app.player.play_next()

    async def show_model(self, model):
        model_type = ModelType(model._meta.model_type)
        if model_type == ModelType.album:
            func = self.show_album
        elif model_type == ModelType.artist:
            func = self.show_artist
        elif model_type == ModelType.playlist:
            func = self.show_playlist
        else:
            def func(model): pass  # seems silly
        self._app.histories.append(model)
        with self._app.create_action('show {}'.format(str(model))):
            await func(model)

    def show_player_playlist(self, songs):
        self.show_songs(songs)
        self.songs_table.song_deleted.connect(
            lambda song: self._app.playlist.remove(song))

    def set_desc(self, desc):
        self._desc_container.show()
        self._desc_container.set_html(desc)

    async def show_playlist(self, playlist):
        self._top_container.show()
        loop = asyncio.get_event_loop()
        songs = await loop.run_in_executor(None, lambda: playlist.songs)
        self._show_songs(songs)
        desc = '<h2>{}</h2>\n{}'.format(playlist.name, playlist.desc or '')
        self.set_desc(desc)
        if playlist.cover:
            loop.create_task(self.show_cover(playlist.cover))

        def remove_song(song):
            model = self.songs_table.model()
            row = model.songs.index(song)
            msg = 'remove {} from {}'.format(song, playlist)
            with self._app.create_action(msg) as action:
                rv = playlist.remove(song.identifier)
                if rv:
                    model.removeRow(row)
                else:
                    action.failed()
        self.songs_table.song_deleted.connect(lambda song: remove_song(song))

    async def show_artist(self, artist):
        self._top_container.show()
        loop = asyncio.get_event_loop()
        future_songs = loop.run_in_executor(None, lambda: artist.songs)
        future_desc = loop.run_in_executor(None, lambda: artist.desc)
        await asyncio.wait([future_songs, future_desc])
        desc = future_desc.result()
        self.set_desc(desc or '<h2>{}</h2>'.format(artist.name))
        self._show_songs(future_songs.result())
        if artist.cover:
            loop.create_task(self.show_cover(artist.cover))

    async def show_album(self, album):
        self._top_container.show()
        loop = asyncio.get_event_loop()
        future_songs = loop.run_in_executor(None, lambda: album.songs)
        future_desc = loop.run_in_executor(None, lambda: album.desc)
        await asyncio.wait([future_songs, future_desc])
        self.set_desc(future_desc.result() or
                      '<h2>{}</h2>'.format(album.name))
        songs = future_songs.result()
        self._show_songs(songs)
        if album.cover:
            loop.create_task(self.show_cover(album.cover))

    def show_collection(self, collection):
        self._top_container.hide()
        self.show_songs(collection.models)
        self.songs_table.song_deleted.connect(collection.remove)

    async def show_url(self, url):
        model = self._app.protocol.get_model(url)
        if model.meta.model_type == ModelType.song:
            self._app.player.play_song(model)
        else:
            # TODO: add artist/album/user support
            self._app.show_msg('暂时只支持歌曲，不支持其它歌曲资源')

    async def show_cover(self, cover):
        # FIXME: cover_hash may not work properly someday
        cover_uid = cover.split('/', -1)[-1]
        content = await self._app.img_ctl.get(cover, cover_uid)
        img = QImage()
        img.loadFromData(content)
        pixmap = QPixmap(img)
        if not pixmap.isNull():
            self.set_cover(pixmap)
            self.update()

    def _show_songs(self, songs):
        try:
            self.songs_table.song_deleted.disconnect()
        except TypeError:  # no connections at all
            pass
        self.show()
        self.songs_table.show()
        songs = songs or []
        logger.debug('Show songs in table, total: %d', len(songs))
        source_name_map = {p.identifier: p.name for p in self._app.library.list()}
        self.songs_table.setModel(SongsTableModel(songs, source_name_map))
        self.songs_table.scrollToTop()

    def show_songs(self, songs):
        self._show_songs(songs)
        self._top_container.show()
        self.hide_desc()
        self.hide_cover()

    def set_cover(self, pixmap):
        self._cover_container.show()
        self._cover_label.setPixmap(
            pixmap.scaledToWidth(self._cover_label.width(),
                                 mode=Qt.SmoothTransformation))

    def toggle_desc_container_fold(self):
        # TODO: add toggle animation?
        if self._desc_container_folded:
            self._top_container.setMaximumHeight(4000)
            self.songs_table.hide()
            self._desc_container_folded = False
        else:
            self._top_container.setMaximumHeight(self.overview_height)
            self.songs_table.show()
            self._desc_container_folded = True

    def search(self, text):
        if self.isVisible() and self.songs_table is not None:
            self.songs_table.filter_row(text)

    def hide_cover(self):
        self._cover_container.hide()

    def hide_desc(self):
        self._desc_container.hide()



import asyncio
import locale
import logging
from functools import partial
from contextlib import contextmanager

from fuocore import LiveLyric, MpvPlayer, Library

from feeluown.protocol import FuoProcotol
from .cliapp import LiveLyricPublisher
from .player import Player
from .plugin import PluginsManager
from .version import VersionManager


logger = logging.getLogger(__name__)


class App(object):
    CliMode = 0x0001
    GuiMode = 0x0010

    mode = 0x0000

    def __init__(self, player_kwargs=None):
        self.player = Player(app=self, **(player_kwargs or {}))
        self.playlist = self.player.playlist
        self.library = Library()
        self.live_lyric = LiveLyric()

        self.protocol = FuoProcotol(self)

        self.plugin_mgr = PluginsManager(self)
        self.version_mgr = VersionManager(self)

    def show_msg(self, msg, *args, **kwargs):
        logger.info(msg)

    def initialize(self):
        self.player.position_changed.connect(self.live_lyric.on_position_changed)
        self.playlist.song_changed.connect(self.live_lyric.on_song_changed)

        self.plugin_mgr.scan()
        loop = asyncio.get_event_loop()
        loop.call_later(10, partial(loop.create_task, self.version_mgr.check_release()))

    @contextmanager
    def create_action(self, s):
        show_msg = self.show_msg

        class Action:
            def set_progress(self, value):
                value = int(value * 100)
                show_msg(s + '...{}%'.format(value), timeout=-1)

            def failed(self):
                show_msg(s + '...failed', timeout=-1)

        show_msg(s + '...', timeout=-1)  # doing
        try:
            yield Action()
        except Exception:
            show_msg(s + '...error')  # error
            raise
        else:
            show_msg(s + '...done')  # done


class CliApp(App):
    mode = App.CliMode

    def __init__(self, pubsub_gateway, player_kwargs=None):
        super().__init__(player_kwargs=player_kwargs)

        self.pubsub_gateway = pubsub_gateway
        self._live_lyric_publisher = LiveLyricPublisher(pubsub_gateway)

        self.live_lyric.sentence_changed.connect(self._live_lyric_publisher.publish)



from enum import Enum
from functools import partial

from PyQt5.QtCore import (
    pyqtSignal,
    QAbstractListModel,
    QAbstractTableModel,
    QModelIndex,
    Qt,
    QSize,
    QTime,
    QVariant,
)
from PyQt5.QtWidgets import (
    QAbstractItemDelegate,
    QAbstractItemView,
    QAction,
    QComboBox,
    QFrame,
    QHBoxLayout,
    QHeaderView,
    QInputDialog,
    QListView,
    QMenu,
    QPushButton,
    QStyledItemDelegate,
    QTableView,
    QWidget,
)

from fuocore.models import ModelExistence
from feeluown.utils import parse_ms
from feeluown.mimedata import ModelMimeData
from feeluown.helpers import use_mac_theme


class Column(Enum):
    song = 0
    source = 1
    duration = 2
    artist = 3
    album = 4


class SongsTableModel(QAbstractTableModel):
    def __init__(self, songs, source_name_map=None):
        super().__init__()
        self.songs = songs
        self._source_set = set()

        # XXX: icon should be a str (charactor symbol)
        self._source_name_map = source_name_map or {}
        self._initialize()

    def _initialize(self):
        for song in self.songs:
            self._source_set.add(song.source)

    def flags(self, index):
        song = index.data(Qt.UserRole)
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable
        if song.exists == ModelExistence.no or index.column() in (0, 2, ):
            return Qt.ItemIsSelectable
        if index.column() == 1:
            return flags | Qt.ItemIsDragEnabled
        return flags

    def rowCount(self, parent=QModelIndex()):
        return len(self.songs)

    def columnCount(self, _):
        return 5

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        sections = ('来源', '歌曲标题', '时长', '歌手', '专辑')
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                if section < len(sections):
                    return sections[section]
                return ''
        else:
            if role == Qt.DisplayRole:
                return section
            elif role == Qt.TextAlignmentRole:
                return Qt.AlignRight
        return QVariant()

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        if index.row() >= len(self.songs) or index.row() < 0:
            return QVariant()

        song = self.songs[index.row()]
        if role in (Qt.DisplayRole, Qt.ToolTipRole):
            if index.column() == 0:
                return self._source_name_map.get(song.source, '').strip()
            elif index.column() == 1:
                return song.title_display
            elif index.column() == 2:
                return song.duration_ms_display
            elif index.column() == 3:
                return song.artists_name_display
            elif index.column() == 4:
                return song.album_name_display
        elif role == Qt.TextAlignmentRole:
            if index.column() == 0:
                return Qt.AlignCenter | Qt.AlignBaseline
        elif role == Qt.EditRole:
            return 1
        elif role == Qt.UserRole:
            return song
        return QVariant()

    def mimeData(self, indexes):
        if len(indexes) == 1:
            index = indexes[0]
            song = index.data(Qt.UserRole)
            return ModelMimeData(song)



class ArtistsModel(QAbstractListModel):
    def __init__(self, artists):
        super().__init__()
        self.artists = artists

    def rowCount(self, parent=QModelIndex()):
        return len(self.artists)

    def data(self, index, role):
        artist = self.artists[index.row()]
        if role == Qt.DisplayRole:
            return artist.name
        elif role == Qt.UserRole:
            return artist
        elif role == Qt.SizeHintRole:
            return QSize(100, 30)
        return QVariant()


class SongOpsEditor(QWidget):
    """song editor for playlist table view"""

    def __init__(self, parent):
        super().__init__(parent)
        self.download_btn = QPushButton('↧', self)
        self.play_btn = QPushButton('☊', self)
        self._layout = QHBoxLayout(self)
        self._layout.addWidget(self.play_btn)
        self._layout.addWidget(self.download_btn)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)


class ArtistsSelectionView(QListView):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.Dialog | Qt.FramelessWindowHint)
        self.setObjectName('artists_selection_view')


class SongsTableDelegate(QStyledItemDelegate):
    def __init__(self, parent):
        super().__init__(parent)
        self.view = parent

    def createEditor(self, parent, option, index):
        # Yeah, I'm a genius, again.
        if index.column() in (2, ):
            editor = SongOpsEditor(parent)
            editor.play_btn.clicked.connect(
                partial(self.closeEditor.emit, editor, QAbstractItemDelegate.SubmitModelCache))
            editor.download_btn.clicked.connect(
                partial(self.closeEditor.emit, editor, QAbstractItemDelegate.SubmitModelCache))
            editor.play_btn.clicked.connect(partial(self.view.play_song_needed.emit, index.data(role=Qt.UserRole)))
            return editor
        elif index.column() == 3:
            editor = ArtistsSelectionView(parent)
            editor.clicked.connect(partial(self.commitData.emit, editor))
            editor.move(parent.mapToGlobal(option.rect.bottomLeft()))
            editor.setFixedWidth(option.rect.width())
            return editor

    def setEditorData(self, editor, index):
        super().setEditorData(editor, index)
        if index.column() == 3:
            song = index.data(role=Qt.UserRole)
            model = ArtistsModel(song.artists)
            editor.setModel(model)
            editor.setCurrentIndex(QModelIndex())

    def setModelData(self, editor, model, index):
        if index.column() == 3:
            index = editor.currentIndex()
            if index.isValid():
                artist = index.data(Qt.UserRole)
                self.view.show_artist_needed.emit(artist)
        super().setModelData(editor, model, index)

    def paint(self, painter, option, index):
        super().paint(painter, option, index)

    def sizeHint(self, option, index):
        widths = (0.1, 0.3, 0.1, 0.2, 0.3)
        width = self.parent().width()
        w = int(width * widths[index.column()])
        return QSize(w, option.rect.height())

    def editorEvent(self, event, model, option, index):
        super().editorEvent(event, model, option, index)
        return False

    def updateEditorGeometry(self, editor, option, index):
        if index.column() != 3:
            super().updateEditorGeometry(editor, option, index)


class SongsTableView(QTableView):

    show_artist_needed = pyqtSignal([object])
    show_album_needed = pyqtSignal([object])
    play_song_needed = pyqtSignal([object])

    # 之后或许可以改成 row_deleted，row_deleted 更抽象，
    # 而 song_deleted 更具体，方便以后修改设计。
    song_deleted = pyqtSignal([object])

    def __init__(self, parent=None):
        super().__init__(parent)

        self.delegate = SongsTableDelegate(self)
        self.setItemDelegate(self.delegate)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        # FIXME: PyQt5 seg fault
        # self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        # macOS 的滚动条可以自动隐藏
        if use_mac_theme():
            self.setFrameShape(QFrame.NoFrame)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().hide()

        self.setMouseTracking(True)
        self.setEditTriggers(QAbstractItemView.SelectedClicked)
        # self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        #self.setFocusPolicy(Qt.NoFocus)
        self.setAlternatingRowColors(True)
        self.setShowGrid(False)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragOnly)
        self.activated.connect(self._on_activated)

    def _on_activated(self, index):
        if index.column() == 1:
            song = index.data(Qt.UserRole)
            self.play_song_needed.emit(song)
        elif index.column() == 3:
            song = index.data(Qt.UserRole)
            artists = song.artists
            if artists is not None:
                if len(artists) > 1:
                    self.edit(index)
                else:
                    self.show_artist_needed.emit(artists[0])
        elif index.column() == 4:
            song = index.data(Qt.UserRole)
            if song.album:
                self.show_album_needed.emit(song.album)
        # FIXME: 在点击之后，音乐数据可能会有更新，理应触发界面更新
        # 测试 dataChanged 似乎不能按照预期工作
        model = self.model()
        topleft = model.createIndex(index.row(), 0)
        bottomright = model.createIndex(index.row(), 4)
        model.dataChanged.emit(topleft, bottomright, [])

    def setModel(self, model):
        super().setModel(model)
        self.show_all_rows()

    def show_all_rows(self):
        for i in range(self.model().rowCount()):
            self.setRowHidden(i, False)

    def filter_row(self, text):
        # TODO: improve search algorithm
        if not text:
            self.show_all_rows()
            return
        if not self.model():
            return

        songs = self.model().songs
        for i, song in enumerate(songs):
            if text.lower() not in song.title_display.lower()\
                    and text not in song.album_name_display.lower()\
                    and text not in song.artists_name_display.lower():
                self.setRowHidden(i, True)
            else:
                self.setRowHidden(i, False)

    def contextMenuEvent(self, event):
        menu = QMenu()
        index = self.indexAt(event.pos())
        self.selectRow(index.row())
        song = self.model().data(index, Qt.UserRole)
        remove_song_action = QAction('移除歌曲', menu)
        remove_song_action.triggered.connect(partial(self.song_deleted.emit, song))
        menu.addAction(remove_song_action)
        menu.exec(event.globalPos())



class Config:
    """配置模块

    用户可以在 rc 文件中配置各个选项的值
    """
    COLLECTIONS_DIR = []

    def __getattr__(self, name):
        if name not in self.__dict__:
            return None
        return object.__getattribute__(self, name)


config = Config()



import asyncio
from functools import partial
import logging
import os

from PyQt5.QtCore import QObject, pyqtSignal

from feeluown.consts import SONG_DIR
from feeluown.utils import emit_requests_progress


logger = logging.getLogger(__name__)


class Downloader(QObject):
    download_progress_signal = pyqtSignal([int])

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.is_downloading = False
        self.current_song = None
        self.queue = []

    def _access_queue(self):
        if self.queue:
            song = self.queue.pop(0)
            self.download_song(song)
        else:
            self._app.message('下载队列下载完毕')

    def download_song(self, song):
        for s in self.queue:
            if s.mid == song.mid:
                self._app.message('%s 已经在下载队列中' % song.title)
                self._app.message('当前正在下载 %s' % self.current_song.title)
                return

        if self.is_downloading:
            if len(self.queue) > 10:
                self._app.message('下载队列已满，请稍后加入', error=True)
                return
            self.queue.append(song)
            self._app.message('%s 加入下载队列之中' % song.title)
            return

        self.is_downloading = True
        event_loop = asyncio.get_event_loop()
        event_loop.create_task(self._download(song))

    @asyncio.coroutine
    def _download(self, song):
        f_name = song.filename
        f_path = os.path.join(SONG_DIR, f_name)
        if os.path.exists(f_path):
            logger.warning('%s have been downloaded' % song.title)
            self._app.message('%s 这首歌已经存在' % song.title)
            self.is_downloading = False
            self._access_queue()
            return True
        self.current_song = song
        if song.url is not None:
            try:
                event_loop = asyncio.get_event_loop()
                self._app.message('准备下载 %s' % song.title)
                res = song._api.http.get(song.url, stream=True, timeout=30)
                if res.status_code != 200:
                    raise Exception('response not ok while download song')
                future = event_loop.run_in_executor(
                    None,
                    partial(emit_requests_progress, res,
                            self.download_progress_signal))
                content = yield from future
                if not content:
                    raise Exception('error in downloaded song')
                with open(f_path, 'wb') as f:
                    f.write(content)
                logger.info('download song %s successfully' % song.title)
                self._app.message('%s 下载成功' % song.title)
                self.is_downloading = False
                self._access_queue()
                return True
            except Exception as e:
                logger.error(e)

        self._app.message('下载 %s 失败' % song.title, error=True)
        self.is_downloading = False
        self._access_queue()
        return False



import logging
import re

from .helpers import (
    ModelType,
    TYPE_NS_MAP,
    URL_SCHEME,
    NS_TYPE_MAP,
    get_url,
)

logger = logging.getLogger(__name__)


class Cmd:
    def __init__(self, action, *args, **kwargs):
        self.action = action
        self.args = args

    def __str__(self):
        return 'action:{} args:{}'.format(self.action, self.args)


class CmdParser:

    @classmethod
    def parse(cls, cmd_str):
        cmd_str = cmd_str.strip()
        cmd_parts = cmd_str.split(' ', 1)
        if not cmd_parts:
            return None
        return Cmd(*cmd_parts)


class ModelParser:
    """
    XXX: 名字叫做 Parser 可能不是很合适？这里可能包含类似 Tokenizor 的功能。
    """

    def __init__(self, library):
        """
        :param library: 音乐库
        """
        self._library = library

    def parse_line(self, line):
        # pylint: disable=too-many-locals
        if not line.startswith('fuo://'):
            return None
        parts = line.split('#')
        if len(parts) == 2:
            url, desc = parts
        else:
            url, desc = parts[0], ''
        source_list = [provider.identifier for provider in self._library.list()]
        ns_list = list(TYPE_NS_MAP.values())
        p = re.compile(r'^fuo://({})/({})/(\w+)$'
                       .format('|'.join(source_list), '|'.join(ns_list)))
        m = p.match(url.strip())
        if not m:
            logger.warning('invalid model url')
            return None
        source, ns, identifier = m.groups()
        provider = self._library.get(source)
        Model = provider.get_model_cls(NS_TYPE_MAP[ns])
        model = None
        if ns == 'songs':
            data = ModelParser.parse_song_desc(desc.strip())
            model = Model.create_by_display(identifier=identifier, **data)
        return model

    def gen_line(self, model):
        """dump model to line"""
        line = get_url(model)
        if model.meta.model_type == ModelType.song:
            desc = self.gen_song_desc(model)
            line += '\t# '
            line += desc
        return line

    @classmethod
    def gen_song_desc(cls, song):
        return '{} - {} - {} - {}'.format(
            song.title_display,
            song.artists_name_display,
            song.album_name_display,
            song.duration_ms_display
        )

    @classmethod
    def parse_song_desc(cls, desc):
        values = desc.split(' - ')
        if len(values) < 4:
            values.extend([''] * (4 - len(values)))
        return {
            'title': values[0],
            'artists_name': values[1],
            'album_name': values[2],
            'duration_ms': values[3]
        }
